"""
Reconstruction algorithms for EIT.

Arithmetic
----------
  - fem: Finite-element method
  - base: EitBase class
  - bp: Back-projection
  - jac: Jacobian matrix based method
  - greit: The GREIT algorithm
  - utils: EIT related helper function
  - interp2d: Spatial interpolation for EIT
"""
